(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontutil.mli,v 1.6 2002/01/28 11:27:53 malekith Exp $
 *
 * Module interface: Gontutil
 * Description: Gont program tree representation + various utilities.
 *)

(* Identifier. *)
type id = string

(* Location in source file. *)
type location = {mutable loc_file: string; 
                 mutable loc_char: int;
                 mutable loc_line: int}

(* Types used by Gont compiler. *)
type ty =
    Int                         (* integer of the same size as pointer *)
  | Void                        (* unit type *)
  | Bool                        (* boolean, same size as Int *)
  | Float                       (* C's double *)
  | String                      (* Pointer to struct string$ *)
  | Tuple of ty list            (* void ** *)
  | Func of ft_record
  | TyVar of id
  | Struct of (ty * id) list
  | OptStruct of (ty * id) list
  | Union of (ty * id) list
  | AbstractType                (* Used when outside module defines `type t' *)
  | NamedType of nt_record      (* <int>List::t, some_struct *)
and typedef = 
   {mutable td_tyvars: id list;  (* Parameters of defined type. *)
    mutable td_type: ty;         (* Actual defined type (Struct or Union in
                                  * most cases). *)
    td_module: id;               (* "" for current module *)
    td_name: id;
    td_loc: location}            (* Where it was defined? *)
and nt_record = 
   {nt_parms: ty list;          (* Actual types to be substituted to
                                 * nt_tydef.ty_vars *)
    nt_module: id;              (* "" if module qualifier not given. *)
    nt_name: id;
    nt_loc: location;           (* Where it was used? We need it, as we might
                                 * want to generate error message about it. *)
    mutable nt_tydef: typedef option; (* Filled by Gontbind. *)
    (* nt_type dropped, use instance_value. *)
   }
(* ft_tyvars is generally filled by the parser using simple_fv function
 * called on parameters and return type. However this is wrong for nested
 * functions (for example `f('a x) { g('a x, 'b y) {} }', ft_tyvars for g
 * should be ['b], but it is ['a, 'b]), this is fixed by Gonttypes module. 
 * That's why we need ft_tyvars to be mutable. *)
and ft_record = {ft_ret_type: ty;
                 ft_parms: ty list;
                 mutable ft_tyvars: id list}
;;

(* Abstract type of symbol. Class conforming it is defined in Gontbind,
 * however it does not export the actuall type outside. *)
class type symbol = 
  object
    method get_name : id
    method get_type : ty
    method set_type : ty -> unit
    method get_fnc_level : int
    method get_scope_level : int
    method in_closure : bool
    method isunion : bool
    method make_union_field : int * typedef -> unit
    method union_type : typedef
    method put_into_closure : unit
    method make_top_pattern : unit
    method is_lvalue : bool
    method put_into_pattern : unit
    method in_pattern : bool
    method field_id : int
  end
  
(* Type of binary operator. Note distinction of Int* and Float* operators.
 * This is not required by the current typechecker, however it probably
 * will if we want type reconstruction. However I'm not sure if it's good
 * idea... *)
type op2 =
    IntPlus
  | IntMinus
  | IntMult
  | IntDiv
  | IntMod
  | FloatPlus
  | FloatMinus
  | FloatMult
  | FloatDiv
  | Assign
  | Eq | Ne | Lt | Gt | Ge | Le
  
(* Type of unary operatrs, Int/Float, as above. *)
type op1 =
    IntNegate
  | FloatNegate

(* Formal parameter of function. parm_symbol set by Gontbind. Note that it
 * still can be in closure (it is assigned there at the function startup). *)
type parm_record = {parm_name: id;
                    parm_type: ty;
                    mutable parm_symbol: symbol option}

(* Expression and statement representation trees. *)
type raw_exp =
    SimpleBinOp of op2 * exp * exp
  | SimpleUnOp of op1 * exp
  | Ref of ref_record   (* reference to variable *)
  | Call of exp * exp list  (* functional variable * parameters *)
  | IntConst of string
  | BoolConst of bool
  (* Integer and Float constants are represented as strings, because
   * we might not be able to support target's integer/float size
   * correctly in Caml (we would have to use Int64 to be reasonably
   * safe, but even this might not be right), and even if, we don't 
   * need the actuall value, we only need to pass it to Ksi compiler
   * (and hope, it will interpret it correctly :^) *)
  | FloatConst of string
  | VoidConst (* () in ML *)
  | StringConst of string
  | TupleExp of exp list
  | UnionExp of union_record
  | Lambda of lambda_record
  (* Module::name *)
  | CrossModuleRef of xmr_record
and exp = {mutable raw_exp: raw_exp;
           (* Filled by Gonttypes, but don't use it directly, call
            * Gonttypes.type_of instead. *)
           mutable exp_type: ty option;
           exp_loc: location}
and ref_record = {
                  (* This should be accessed only by Gontbind, after binding
                   * you should use ref_symbol#get_name (as it includes
                   * module name for global variables, and might be mangled
                   * for locals). *)
                  ref_name: id;
                  (* How deep in closure chain should we go to get to
                   * the value. If ref_level == 0, the value is defined
                   * at the same function level, and we don't need to
                   * go anywhere to get to it :^), however if 
                   * ref_symbol#in_closure, we need to access it through
                   * cl$ pointer. *)
                  mutable ref_level: int;
                  (* Symbol filled by Gontbind. *)
                  mutable ref_symbol: symbol option}
and xmr_record = {
                  (* Same as for ref_record.ref_name -- don't use after 
                   * binding. *)
                  xmr_module: id;
                  xmr_name: id;
                  (* Symbol filled by Gontbind. *)
                  mutable xmr_symbol: symbol option}
and union_record = {
                    (* Actually it is member of union, one record for each
                       member *)
                    ur_exp: exp;
                    ur_module: id;
                    ur_name: id;
                    mutable ur_sym: symbol option}
and raw_stmt =
    Skip
  | Exp of exp                  (* Expression statement -- evaluate expression
                                 * and forget the value. *)
  | Return of exp               (* Return given value from function. *)
  | ReturnVoid                  (* Hm.. we have currently no way of 
                                 * representing void value... (i.e. '()') *)
  | VarDef of var_record
  | Block of stmt list
  | Pair of stmt * stmt         (* Doesn't introduce new scope, simply
                                 * groups statements, it is introduced
                                 * by few parser rules, however there is
                                 * no user-mode way to get it. *)
  (* Stores base expression which is matched to first fitting pattern
     in (pattern*stmt) list and its stmt is executed *)
  | Switch of exp * (case_clouse list)
  | If of exp * stmt * stmt
  (* All loops are constructed from these two terms. for(a;b;c)X loop is
   * splited into Block(a,Loop(Pair(Pair(BreakIfNot(b),X),c))), while(c)X
   * into Loop(Pair(BreakIfNot(c),X)), and so on. *)
  | Loop of stmt
  | BreakIfNot of exp
  (* Local open -- import all symbols from given module into current 
   * namespace, however note, symbols local to module always override
   * external symbols! *)
  | Open of id
and stmt = {raw_stmt: raw_stmt;
            stmt_loc: location}
and var_record = {var_name: id;         (* Don't use after binding. *)
                  var_type: ty;
                  (* Filled by Gontbind. *)
                  mutable var_symbol: symbol option}
and lambda_record = {lambda_ret_type: ty;
                     lambda_parms: parm_record list;
                     lambda_body: stmt;
                     lambda_tyvars: id list}    (* Free type variables. *)
and case_clouse = {cc_pattern: pattern;
                   cc_stmt: stmt}
and pattern = {mutable raw_pattern: raw_pattern}
(* Type of pattern in pattern matching statements, Tuple/Union so far *)
and raw_pattern = 
    VarPattern of var_pattern_record    
  | ExprPattern of exp
  | TuplePattern of tuple_pattern_record
  | UnionPattern of union_pattern_record
and tuple_pattern_record = {tpr_patterns: pattern list;
                            tpr_loc: location}
and union_pattern_record = {upr_module: id;
                            upr_name: id;
                            (* Filled by Gontbind. *)
                            mutable upr_sym: symbol option;
                            upr_pattern: pattern;
                            upr_loc: location}
and var_pattern_record = {vp_name: id;     (* Don't use after binding. *)
                          (* Filled by Gontbind. *)
                          mutable vp_symbol: symbol option;
                          vp_loc: location}

type fn_decl = {fn_ret_type: ty; 
                fn_name: id; 
                fn_parms: parm_record list;
                fn_body: stmt;
                fn_loc: location;
                fn_tyvars: id list;
                mutable fn_symbol: symbol option;
                fn_export_as: string}

type extern_c_decl = {xc_type: ty; 
                      xc_name: id; 
                      xc_loc: location;
                      mutable xc_symbol: symbol option}

type global_decl = {global_type: ty;
                    global_loc: location;
                    global_name: id;
                    mutable global_symbol: symbol option}

type fn_proto_decl = {fp_type: ty;
                      fp_name: id;
                      fp_loc: location;
                      mutable fp_symbol: symbol option}

type decl =
    FuncDecl of fn_decl
  | GlobalDecl of global_decl
  | ExternC of extern_c_decl
  | FuncProto of fn_proto_decl
  | TopLevelOpen of id
  | TypeDef of typedef
;;

(* Oh well... this is also used for type errors and so on... possibly
 * name should be changed... *)
exception Syntax_error of location * string

(* Some general error, raised when we don't know the actuall location. *)
exception Error of string

(* Go figure... *)
exception Eof
;;

(* Utils. *)

(* Reset compiler state. This generally just resets seqence counters. *)
val reset : unit -> unit

(* Return unique name prefixed with string. *)
val seq_name : id -> id

(* Output debugging (trace) message. *)
val debug : string -> unit

(* An alias for Printf.sprintf. *)
val xf : ('a, unit, string) format -> 'a

(* Display warning. *)
val warn : location * string -> unit

(* Display error message. *)
val err_msg : location * string -> unit

(* Raise Syntax_error. *)
val err : location * string -> 'a

(* 'a is type for context, 'b is input value, and 'c is output value. *)
val list_apply : ('a * 'b -> 'a * 'c) -> 'a * 'b list -> 'a * 'c list

(* Apply given function to type recursivly. Function should not call
 * type_walk. *)
val type_walk : (ty -> ty) -> ty -> ty

(* Return list of free type variables in type. It does not go into
 * named types. *)
val simple_fv : ty -> id list

(* Make function type with tyvars set with simple_fv. Utility used by
 * Parser and Gontbind modules. *)
val make_fn_type : (ty * (parm_record list)) -> ty
;;
