(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: main.ml,v 1.2 2002/01/28 11:28:45 malekith Exp $
 *
 * Module implementation: Main
 * Description: Driver for Gont compiler.
 *)

let compile_only = ref false
let generate_ksi_only = ref false
let save_temps = ref false
let ksi_compiler = ref "xgcc"
let ksi_opts : string list ref = ref []
let input_files : string list ref = ref []
let verbose_flag = ref false
let output_file = ref "a.out"
let prefix = ref Config.prefix
;;

let verb s = if !verbose_flag then (print_endline s; flush stdout) else ()
let err s = prerr_endline ((Array.get Sys.argv 0) ^ ": " ^ s); flush stderr
;;

let xf = Printf.sprintf
;;

let split_filename s =
  let (dir, file) = (Filename.dirname s, Filename.basename s) in
  let base = 
    try
      Filename.chop_extension file
    with Invalid_argument _ -> file in
  let ext = String.sub file ((String.length base) + 1)
                            (String.length file - String.length base - 1) in
  (dir, base, ext)
;;

let execute argv =
  let cmd = String.concat " " argv in
  verb (xf "executing `%s'" cmd);
  let ret = Sys.command cmd in
  if ret != 0 then
    (err (xf "`%s' failed, exit code = %d" cmd ret);
     exit 1)
  else
    ()
;;

let compile_file f = 
  let compile_ksi i o =
    verb (xf "compiling Ksi file `%s' -> `%s'" i o);
    execute ((!ksi_compiler :: !ksi_opts) @ [ "-o"; o; "-c"; i ]) in
  let compile_gont iface i o m =
    let info = {Gont.ci_target_prefix = !prefix;
                Gont.ci_input_file = i;
                Gont.ci_output_file = o;
                Gont.ci_module_name = m} in
    verb (xf "compiling %sGont file `%s' -> `%s'" 
             (if iface then "iface " else "") i o);
    let e = 
      if iface then 
        Gont.compile_iface info
      else
        Gont.compile info in
    if e != 0 then
       (if (not !save_temps) && (Sys.file_exists o) then
          Sys.remove o
        else
          ();
        exit 1) 
     else 
       () in
  let (dir, basename, suffix) = split_filename f in
    let predir = Filename.concat dir in
    match suffix with
        "g" -> 
          let aux = predir (basename ^ ".ksi") in
          let out = predir (basename ^ ".o") in
          let m = String.capitalize basename in
          compile_gont false f aux m;
          if !generate_ksi_only then
            aux
          else
            (compile_ksi aux out; if not !save_temps then Sys.remove aux; out)
      | "o" -> f
      | "ksi" ->
          if !generate_ksi_only then
            f
          else
            let out = predir (basename ^ ".o") in
            (compile_ksi f out; out)
      | "gi" ->
          let out = predir (basename ^ ".gio") in
          let m = String.capitalize basename in
          compile_gont true f out m; out
      | _ ->
          err (xf "don't know what to do with `%s'" f);
          exit 1
;;

let link fs =
  let o = !output_file in
  verb (xf "linking `%s' -> `%s'" (String.concat "', `" fs) o);
  execute ((!ksi_compiler :: !ksi_opts) @ 
          [ "-o"; o; ] @ 
          fs @ 
          [Filename.concat !prefix "stdlib.o"])
;;

open Arg

let main () =
  let add_ksi_opt x = 
     ksi_opts := x :: !ksi_opts in
  let add_input_file x =
     input_files := x :: !input_files in
  let prog = (Array.get Sys.argv 0) in
  let help_str = 
        prog ^ " - compile and link Gont files\n\n\
        If input file name matches *.g, it is assumed to be Gont source\n\
        file, is compiled to *.ksi file, which is in turn, passed to Ksi\n\
        compiler. *.ksi files are passed directly to Ksi compiler.\n" in
  let opt = [
    ("-c", Set compile_only, "         just compile");
    ("-save-temps", Set save_temps, "save temporary files");
    ("-ksi", String add_ksi_opt, "OPT    pass option to ksi compiler");
    ("-ksicc", String (fun s -> ksi_compiler := s), "PROG set Ksi compiler");
    ("-O0", Unit (fun () -> add_ksi_opt "-O0"), "        same as -ksi -O0");
    ("-O1", Unit (fun () -> add_ksi_opt "-O1"), "        same as -ksi -O1");
    ("-O2", Unit (fun () -> add_ksi_opt "-O2"), "        same as -ksi -O2");
    ("-O3", Unit (fun () -> add_ksi_opt "-O3"), "        same as -ksi -O3");
    ("-g", Unit (fun () -> add_ksi_opt "-g"), "         same as -ksi -g");
    ("-S", Set generate_ksi_only, "         just generate .ksi file and stop");
    ("-o", String (fun s -> output_file := s), "OUTFILE  set linker output \
                                                file name");
    ("-B", String (fun s -> prefix := s ), "PREFIX   overide prefix (default " 
                                           ^ !prefix ^ ")");
    ("-v", Set verbose_flag, "         be verbose");
    ("-dump", String (fun s -> exit(Gont.dump_module s)), "MODULE  dump named \
                                                           module and exit")
  ] in

  Arg.parse opt add_input_file help_str;
  ksi_opts := List.rev !ksi_opts;
  (match !input_files with
      [] -> err "so what?";
            err ("run " ^ prog ^ " -help to get some info");
            exit 1
    | _ -> ());
  let out = List.map compile_file (List.rev !input_files) in
    if !compile_only || !generate_ksi_only then
      ()
    else
      link out;
  exit 0
;;

main ()
;;
